﻿using System;
using System.Runtime.ConstrainedExecution;
using System.Runtime.InteropServices;
using System.Text;

namespace TestProject
{
    /// <summary>
    /// API 헬퍼
    /// </summary>
    public static class APIHelper
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Import
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region NT 시스템 정보 질의하기 - NtQuerySystemInformation(systemInformationClass, systemInformation, systemInformationLength, returnLength)

        /// <summary>
        /// NT 시스템 정보 질의하기
        /// </summary>
        /// <param name="systemInformationClass">시스템 정보 클래스</param>
        /// <param name="systemInformation">시스템 정보</param>
        /// <param name="systemInformationLength">시스템 정보 길이</param>
        /// <param name="returnLength">반환 길이</param>
        /// <returns>NT 상태</returns>
        [DllImport("ntdll.dll")]
        public static extern NTStatus NtQuerySystemInformation([In] SystemInformationClass systemInformationClass, [In] IntPtr systemInformation, [In] int systemInformationLength, [Out] out int returnLength);

        #endregion
        #region NT 객체 질의하기 - NtQueryObject(handle, objectInformationClass, objectInformation, objectInformationLength, returnLength)

        /// <summary>
        /// NT 객체 질의하기
        /// </summary>
        /// <param name="handle">핸들</param>
        /// <param name="objectInformationClass">객체 정보 클래스</param>
        /// <param name="objectInformation">객체 정보</param>
        /// <param name="objectInformationLength">객체 정보 길이</param>
        /// <param name="returnLength">반환 길이</param>
        /// <returns>처리 결과</returns>
        [DllImport("ntdll.dll")]
        public static extern NTStatus NtQueryObject([In] IntPtr handle, [In] ObjectInformationClass objectInformationClass, [In] IntPtr objectInformation, [In] int objectInformationLength, [Out] out int returnLength);

        #endregion
        #region 프로세스 열기 - OpenProcess(desiredAccess, inheritHandle, processID)

        /// <summary>
        /// 프로세스 열기
        /// </summary>
        /// <param name="desiredAccess">희망 액세스</param>
        /// <param name="inheritHandle">상속 핸들</param>
        /// <param name="processID">프로세스 ID</param>
        /// <returns>안전 프로세스 핸들</returns>
        [DllImport("kernel32.dll", SetLastError = true)]
        public static extern SafeProcessHandle OpenProcess([In] ProcessAccessRights desiredAccess, [In, MarshalAs(UnmanagedType.Bool)] bool inheritHandle, [In] int processID);

        #endregion
        #region 핸들 복제하기 - DuplicateHandle(sourceProcessHandle, sourceHandle, targetProcessHandle, targetHandle, desiredAccess, inheritHandle, option)

        /// <summary>
        /// 핸들 복제하기
        /// </summary>
        /// <param name="sourceProcessHandle">소스 프로세스 핸들</param>
        /// <param name="sourceHandle">소스 핸들</param>
        /// <param name="targetProcessHandle">타겟 프로세스 핸들</param>
        /// <param name="targetHandle">타겟 핸들</param>
        /// <param name="desiredAccess">희망 액세스</param>
        /// <param name="inheritHandle">상속 핸들</param>
        /// <param name="option">옵션</param>
        /// <returns>처리 결과</returns>
        [DllImport("kernel32.dll", SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        public static extern bool DuplicateHandle([In] IntPtr sourceProcessHandle, [In] IntPtr sourceHandle, [In] IntPtr targetProcessHandle, [Out] out SafeObjectHandle targetHandle, [In] int desiredAccess, [In, MarshalAs(UnmanagedType.Bool)] bool inheritHandle, [In] DuplicateHandleOptions option);

        #endregion
        #region 현재 프로세스 구하기 - GetCurrentProcess()

        /// <summary>
        /// 현재 프로세스 구하기
        /// </summary>
        /// <returns>프로세스 핸들</returns>
        [DllImport("kernel32.dll")]
        public static extern IntPtr GetCurrentProcess();

        #endregion
        #region 프로세스 ID 구하기 - GetProcessId(Process)

        /// <summary>
        /// 프로세스 ID 구하기
        /// </summary>
        /// <param name="Process">프로세스</param>
        /// <returns>프로세스 ID</returns>
        [DllImport("kernel32.dll", SetLastError = true)]
        public static extern int GetProcessId([In] IntPtr Process);

        #endregion
        #region 핸들 닫기 - CloseHandle(objectHandle)

        /// <summary>
        /// 핸들 닫기
        /// </summary>
        /// <param name="objectHandle">객체 핸들</param>
        /// <returns>처리 결과</returns>
        [ReliabilityContract(Consistency.WillNotCorruptState, Cer.Success)]
        [DllImport("kernel32.dll", SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        public static extern bool CloseHandle([In] IntPtr objectHandle);

        #endregion
        #region 도스 장치 질의하기 - QueryDosDevice(deviceName, targetPath, maximumCount)

        /// <summary>
        /// 도스 장치 질의하기
        /// </summary>
        /// <param name="deviceName">장치명</param>
        /// <param name="targetPath">타겟 경로</param>
        /// <param name="maximumCount">최대 수</param>
        /// <returns>처리 결과</returns>
        [DllImport("kernel32.dll", SetLastError = true)]
        public static extern int QueryDosDevice([In] string deviceName, [Out] StringBuilder targetPath, [In] int maximumCount);

        #endregion
    }
}