﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;

namespace TestProject
{
    /// <summary>
    /// 오픈 파일 헬퍼
    /// </summary>
    public class OpenFileHelper
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 장치 딕셔너리
        /// </summary>
        private static Dictionary<string, string> _deviceDictionary;

        /// <summary>
        /// 핸들 타입 토큰 배열
        /// </summary>
        private static readonly string[] _handleTypeTokenArray = new string[]
        { 
            "",
            "",
            "Directory",
            "SymbolicLink",
            "Token",
            "Process",
            "Thread",
            "Unknown7",
            "Event",
            "EventPair",
            "Mutant",
            "Unknown11",
            "Semaphore",
            "Timer",
            "Profile",
            "WindowStation",
            "Desktop",
            "Section",
            "Key",
            "Port",
            "WaitablePort",
            "Unknown21",
            "Unknown22",
            "Unknown23",
            "Unknown24", 
            "IoCompletion",
            "File"
        };

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 네트워크 장치 접두사
        /// </summary>
        private const string NETWORK_DEVICE_PREFIX = "\\Device\\LanmanRedirector\\";

        /// <summary>
        /// 최대 경로
        /// </summary>
        private const int MAXIMUM_PATH = 260;

        /// <summary>
        /// 핸들 타입 토큰 수
        /// </summary>
        private const int HANDLE_TYPE_TOKEN_COUNT = 27;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 오픈 파일 열거자 구하기 - GetOpenFileEnumerator(processID)

        /// <summary>
        /// 오픈 파일 열거자 구하기
        /// </summary>
        /// <param name="processID">프로세스 ID</param>
        /// <returns>파일 시스템 정보 열거자</returns>
        public static IEnumerator<FileSystemInfo> GetOpenFileEnumerator(int processID)
        {
            try
            {
                return new OpenHandle(processID).GetFileEnumerator();
            }
            catch
            {
            }

            return null;
        }

        #endregion
        #region 오픈 핸들 열거자 구하기 - GetOpenHandleEnumerator(processID)

        /// <summary>
        /// 오픈 핸들 열거자 구하기
        /// </summary>
        /// <param name="processID">프로세스 ID</param>
        /// <returns>핸들 정보 열거자</returns>
        public static IEnumerator<HandleInfo> GetOpenHandleEnumerator(int processID)
        {
            try
            {
                return new OpenHandle(processID).GetHandleEnumerator();
            }
            catch
            {
            }

            return null;
        }

        #endregion

        #region 핸들에서 파일명 구하기 - GetFileNameFromHandle(handle, processID, fileName)

        /// <summary>
        /// 핸들에서 파일명 구하기
        /// </summary>
        /// <param name="handle">핸들</param>
        /// <param name="processID">프로세스 ID</param>
        /// <param name="fileName">파일명</param>
        /// <returns>처리 결과</returns>
        public static bool GetFileNameFromHandle(IntPtr handle, int processID, out string fileName)
        {
            IntPtr currentProcessHandle = APIHelper.GetCurrentProcess();

            bool isRemote = (processID != APIHelper.GetProcessId(currentProcessHandle));

            SafeProcessHandle processHandle = null;
            SafeObjectHandle  objectHandle  = null;

            try
            {
                if(isRemote)
                {
                    processHandle = APIHelper.OpenProcess(ProcessAccessRights.ProcessDuplicateHandle, true, processID);

                    if(APIHelper.DuplicateHandle(processHandle.DangerousGetHandle(), handle, currentProcessHandle, out objectHandle, 0, false, DuplicateHandleOptions.SameAccess))
                    {
                        handle = objectHandle.DangerousGetHandle();
                    }
                }

                return GetFileNameFromHandle(handle, out fileName, 200);
            }
            finally
            {
                if(isRemote)
                {
                    if(processHandle != null)
                    {
                        processHandle.Close();
                    }

                    if(objectHandle != null)
                    {
                        objectHandle.Close();
                    }
                }
            }
        }

        #endregion
        #region 핸들에서 파일명 구하기 - GetFileNameFromHandle(handle, fileName)

        /// <summary>
        /// 핸들에서 파일명 구하기
        /// </summary>
        /// <param name="handle">핸들</param>
        /// <param name="fileName">파일명</param>
        /// <returns>처리 결과</returns>
        public static bool GetFileNameFromHandle(IntPtr handle, out string fileName)
        {
            IntPtr pointer = IntPtr.Zero;

            RuntimeHelpers.PrepareConstrainedRegions();

            try
            {
                int length = 0x200;

                RuntimeHelpers.PrepareConstrainedRegions();

                try
                {
                }
                finally
                {
                    pointer = Marshal.AllocHGlobal(length);
                }

                NTStatus status = APIHelper.NtQueryObject(handle, ObjectInformationClass.NameInformation, pointer, length, out length);

                if(status == NTStatus.BufferOverflow)
                {
                    RuntimeHelpers.PrepareConstrainedRegions();

                    try
                    {
                    }
                    finally
                    {
                        Marshal.FreeHGlobal(pointer);

                        pointer = Marshal.AllocHGlobal(length);
                    }

                    status = APIHelper.NtQueryObject(handle, ObjectInformationClass.NameInformation, pointer, length, out length);
                }

                if(status == NTStatus.Success)
                {
                    if(IntPtr.Size == 4)
                    {
                        fileName = Marshal.PtrToStringUni((IntPtr)((int)pointer + 8), (length - 9) / 2);
                    }
                    else
                    {
                        long position = pointer.ToInt64();

                        IntPtr namePointer = new IntPtr(pointer.ToInt64() + 16);

                        fileName = Marshal.PtrToStringUni(namePointer);
                    }

                    return fileName.Length != 0;
                }
            }
            finally
            {
                Marshal.FreeHGlobal(pointer);
            }

            fileName = string.Empty;

            return false;
        }

        #endregion
        #region 핸들에서 파일명 구하기 - GetFileNameFromHandle(state)

        /// <summary>
        /// 핸들에서 파일명 구하기
        /// </summary>
        /// <param name="state">상태</param>
        public static void GetFileNameFromHandle(object state)
        {
            FileNameFromHandleState fileNameFromHandleState = state as FileNameFromHandleState;

            string fileName;

            fileNameFromHandleState.ReturnValue = GetFileNameFromHandle(fileNameFromHandleState.Handle, out fileName);
            fileNameFromHandleState.FileName = fileName;

            fileNameFromHandleState.Set();
        }

        #endregion
        #region 핸들에서 파일명 구하기 - GetFileNameFromHandle(handle, fileName, wait)

        /// <summary>
        /// 핸들에서 파일명 구하기
        /// </summary>
        /// <param name="handle">핸들</param>
        /// <param name="fileName">파일명</param>
        /// <param name="wait">대기</param>
        /// <returns>처리 결과</returns>
        public static bool GetFileNameFromHandle(IntPtr handle, out string fileName, int wait)
        {
            using(FileNameFromHandleState state = new FileNameFromHandleState(handle))
            {
                ThreadPool.QueueUserWorkItem(new WaitCallback(GetFileNameFromHandle), state);

                if(state.WaitOne(wait))
                {
                    fileName = state.FileName;

                    return state.ReturnValue;
                }
                else
                {
                    fileName = string.Empty;

                    return false;
                }
            }
        }

        #endregion

        #region 핸들 타입 구하기 - GetHandleType(handle, processID, handleType)

        /// <summary>
        /// 핸들 타입 구하기
        /// </summary>
        /// <param name="handle">핸들</param>
        /// <param name="processID">프로세스 ID</param>
        /// <param name="handleType">시스템 핸들 타입</param>
        /// <returns>처리 결과</returns>
        public static bool GetHandleType(IntPtr handle, int processID, out SystemHandleType handleType)
        {
            string token = GetHandleTypeToken(handle, processID);

            return GetHandleTypeFromToken(token, out handleType);
        }

        #endregion
        #region 핸들 타입 구하기 - GetHandleType(handle, handleType)

        /// <summary>
        /// 핸들 타입 구하기
        /// </summary>
        /// <param name="handle">핸들</param>
        /// <param name="handleType">시스템 핸들 타입</param>
        /// <returns>처리 결과</returns>
        public static bool GetHandleType(IntPtr handle, out SystemHandleType handleType)
        {
            string token = GetHandleTypeToken(handle);

            return GetHandleTypeFromToken(token, out handleType);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Private

        #region 토큰에서 핸들 타입 구하기 - GetHandleTypeFromToken(token, handleType)

        /// <summary>
        /// 토큰에서 핸들 타입 구하기
        /// </summary>
        /// <param name="token">토큰</param>
        /// <param name="handleType">시스템 핸들 타입</param>
        /// <returns>처리 결과</returns>
        private static bool GetHandleTypeFromToken(string token, out SystemHandleType handleType)
        {
            for(int i = 1; i < HANDLE_TYPE_TOKEN_COUNT; i++)
            {
                if(_handleTypeTokenArray[i] == token)
                {
                    handleType = (SystemHandleType)i;

                    return true;
                }
            }

            handleType = SystemHandleType.Unknown;

            return false;
        }

        #endregion

        #region 핸들 타입 토큰 구하기 - GetHandleTypeToken(handle)

        /// <summary>
        /// 핸들 타입 토큰 구하기
        /// </summary>
        /// <param name="handle">핸들</param>
        /// <returns>핸들 타입 토큰</returns>
        private static string GetHandleTypeToken(IntPtr handle)
        {
            int length;

            APIHelper.NtQueryObject(handle, ObjectInformationClass.TypeInformation, IntPtr.Zero, 0, out length);

            if(length <= 0)
            {
                return string.Empty;
            }

            IntPtr pointer = IntPtr.Zero;

            RuntimeHelpers.PrepareConstrainedRegions();

            try
            {
                RuntimeHelpers.PrepareConstrainedRegions();

                try
                {
                }
                finally
                {
                    pointer = Marshal.AllocHGlobal(length);
                }

                if(APIHelper.NtQueryObject(handle, ObjectInformationClass.TypeInformation, pointer, length, out length) == NTStatus.Success)
                {
                    if(IntPtr.Size == 4)
                    {
                        return Marshal.PtrToStringUni((IntPtr)((int)pointer + 0x60));
                    }
                    else
                    {
                        IntPtr namePointer = new IntPtr(pointer.ToInt64() + 104);

                        return Marshal.PtrToStringUni(namePointer);
                    }
                }
            }
            finally
            {
                Marshal.FreeHGlobal(pointer);
            }

            return string.Empty;
        }

        #endregion
        #region 핸들 타입 토큰 구하기 - GetHandleTypeToken(handle, processID)

        /// <summary>
        /// 핸들 타입 토큰 구하기
        /// </summary>
        /// <param name="handle">핸들</param>
        /// <param name="processID">프로세스 ID</param>
        /// <returns>핸들 타입 토큰</returns>
        private static string GetHandleTypeToken(IntPtr handle, int processID)
        {
            IntPtr currentProcessHandle = APIHelper.GetCurrentProcess();

            bool isRemote = (processID != APIHelper.GetProcessId(currentProcessHandle));

            SafeProcessHandle processHandle = null;
            SafeObjectHandle  objectHandle  = null;
            try
            {
                if(isRemote)
                {
                    processHandle = APIHelper.OpenProcess(ProcessAccessRights.ProcessDuplicateHandle, true, processID);

                    if(APIHelper.DuplicateHandle(processHandle.DangerousGetHandle(), handle, currentProcessHandle, out objectHandle, 0, false, DuplicateHandleOptions.SameAccess))
                    {
                        handle = objectHandle.DangerousGetHandle();
                    }
                }

                return GetHandleTypeToken(handle);
            }
            finally
            {
                if(isRemote)
                {
                    if(processHandle != null)
                    {
                        processHandle.Close();
                    }

                    if(objectHandle != null)
                    {
                        objectHandle.Close();
                    }
                }
            }
        }

        #endregion

        #region 장치명 정규화하기 - NormalizeDeviceName(deviceName)

        /// <summary>
        /// 장치명 정규화하기
        /// </summary>
        /// <param name="deviceName">장치명</param>
        /// <returns>정규화 장치명</returns>
        private static string NormalizeDeviceName(string deviceName)
        {
            if(string.Compare(deviceName, 0, NETWORK_DEVICE_PREFIX, 0, NETWORK_DEVICE_PREFIX.Length, StringComparison.InvariantCulture) == 0)
            {
                string shareName = deviceName.Substring(deviceName.IndexOf('\\', NETWORK_DEVICE_PREFIX.Length) + 1);

                return string.Concat(NETWORK_DEVICE_PREFIX, shareName);
            }

            return deviceName;
        }

        #endregion
        #region 장치 딕셔너리 만들기 - BuildDeviceDictionary()

        /// <summary>
        /// 장치 딕셔너리 만들기
        /// </summary>
        /// <returns>장치 딕셔너리</returns>
        private static Dictionary<string, string> BuildDeviceDictionary()
        {
            string[] logicalDriveArray = Environment.GetLogicalDrives();

            Dictionary<string, string> localDeviceDictionary = new Dictionary<string, string>(logicalDriveArray.Length);

            StringBuilder stringBuilder = new StringBuilder(MAXIMUM_PATH);

            foreach(string drive in logicalDriveArray)
            {
                string deviceName = drive.Substring(0, 2);

                APIHelper.QueryDosDevice(deviceName, stringBuilder, MAXIMUM_PATH);

                localDeviceDictionary.Add(NormalizeDeviceName(stringBuilder.ToString()), deviceName);
            }

            localDeviceDictionary.Add(NETWORK_DEVICE_PREFIX.Substring(0, NETWORK_DEVICE_PREFIX.Length - 1), "\\");

            return localDeviceDictionary;
        }

        #endregion
        #region 장치 딕셔너리 보장하기 - EnsureDeviceDictionary()

        /// <summary>
        /// 장치 딕셔너리 보장하기
        /// </summary>
        private static void EnsureDeviceDictionary()
        {
            if(_deviceDictionary == null)
            {
                Dictionary<string, string> localDeviceDictionary = BuildDeviceDictionary();

                Interlocked.CompareExchange<Dictionary<string, string>>(ref _deviceDictionary, localDeviceDictionary, null);
            }
        }

        #endregion

        #region 장치 경로를 DOS 경로로 변환하기 - ConvertDevicePathToDosPath(devicePath, dosPath)

        /// <summary>
        /// 장치 경로를 DOS 경로로 변환하기
        /// </summary>
        /// <param name="devicePath">장치 경로</param>
        /// <param name="dosPath">DOS 경로</param>
        /// <returns>처리 결과</returns>
        public static bool ConvertDevicePathToDosPath(string devicePath, out string dosPath)
        {
            EnsureDeviceDictionary();

            int length = devicePath.Length;

            while(length > 0 && (length = devicePath.LastIndexOf('\\', length - 1)) != -1)
            {
                string drive;

                if(_deviceDictionary.TryGetValue(devicePath.Substring(0, length), out drive))
                {
                    dosPath = string.Concat(drive, devicePath.Substring(length));

                    return dosPath.Length != 0;
                }
            }

            dosPath = string.Empty;

            return false;
        }

        #endregion
    }
}