﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;

namespace TestProject
{
    /// <summary>
    /// 오픈 핸들
    /// </summary>
    public sealed class OpenHandle
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 프로세스 OD
        /// </summary>
        private readonly int processID;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - OpenHandle(processID)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="processID">프로세스 ID</param>
        public OpenHandle(int processID)
        {
            this.processID = processID;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 파일 열거자 구하기 - GetFileEnumerator()

        /// <summary>
        /// 파일 열거자 구하기
        /// </summary>
        /// <returns>파일 시스템 정보 열거자</returns>
        public IEnumerator<FileSystemInfo> GetFileEnumerator()
        {
            NTStatus status;

            int length = 0x10000;

            do
            {
                IntPtr pointer = IntPtr.Zero;

                RuntimeHelpers.PrepareConstrainedRegions();

                try
                {
                    RuntimeHelpers.PrepareConstrainedRegions();

                    try
                    {
                    }
                    finally
                    {
                        pointer = Marshal.AllocHGlobal(length);
                    }

                    int returnLength;

                    status = APIHelper.NtQuerySystemInformation(SystemInformationClass.HandleInformation, pointer, length, out returnLength);

                    if(status == NTStatus.InfoLengthMismatch)
                    {
                        length = ((returnLength + 0xffff) & ~0xffff);
                    }
                    else if(status == NTStatus.Success)
                    {
                        long handleCount = 0;

                        if(IntPtr.Size == 4)
                        {
                            handleCount = Marshal.ReadInt32(pointer);
                        }
                        else
                        {
                            handleCount = Marshal.ReadInt64(pointer);
                        }

                        int offset = IntPtr.Size;

                        int size = Marshal.SizeOf(typeof(SystemHandleEntry));

                        for(int i = 0; i < handleCount; i++)
                        {
                            SystemHandleEntry handleEntry = (SystemHandleEntry)Marshal.PtrToStructure((IntPtr)(pointer.ToInt64() + offset), typeof(SystemHandleEntry));

                            IntPtr newPointer = new IntPtr(pointer.ToInt64() + offset);

                            int newProcessID = Marshal.ReadInt32(newPointer);

                            if(handleEntry.OwnerProcessID == processID)
                            {
                                IntPtr handle = (IntPtr)handleEntry.HandleValue;

                                SystemHandleType handleType;

                                if(OpenFileHelper.GetHandleType(handle, handleEntry.OwnerProcessID, out handleType) && handleType == SystemHandleType.File)
                                {
                                    string devicePath;

                                    if(OpenFileHelper.GetFileNameFromHandle(handle, handleEntry.OwnerProcessID, out devicePath))
                                    {
                                        string dosPath;

                                        if(OpenFileHelper.ConvertDevicePathToDosPath(devicePath, out dosPath))
                                        {
                                            if(File.Exists(dosPath))
                                            {
                                                yield return new FileInfo(dosPath);
                                            }
                                            else if(Directory.Exists(dosPath))
                                            {
                                                yield return new DirectoryInfo(dosPath);
                                            }
                                        }
                                    }
                                }
                            }

                            offset += size;
                        }
                    }
                }
                finally
                {
                    Marshal.FreeHGlobal(pointer);
                }
            }
            while(status == NTStatus.InfoLengthMismatch);
        }

        #endregion
        #region 핸들 열거자 구하기 - GetHandleEnumerator()

        /// <summary>
        /// 핸들 열거자 구하기
        /// </summary>
        /// <returns>핸들 열거자</returns>
        public IEnumerator<HandleInfo> GetHandleEnumerator()
        {
            NTStatus status;

            int length = 0x10000;

            do
            {
                IntPtr pointer = IntPtr.Zero;

                RuntimeHelpers.PrepareConstrainedRegions();

                try
                {
                    RuntimeHelpers.PrepareConstrainedRegions();

                    try
                    {
                    }
                    finally
                    {
                        pointer = Marshal.AllocHGlobal(length);
                    }

                    int returnLength;

                    status = APIHelper.NtQuerySystemInformation(SystemInformationClass.HandleInformation, pointer, length, out returnLength);

                    if(status == NTStatus.InfoLengthMismatch)
                    {
                        length = ((returnLength + 0xffff) & ~0xffff);
                    }
                    else if(status == NTStatus.Success)
                    {
                        long handleCount = 0;

                        if(IntPtr.Size == 4)
                        {
                            handleCount = Marshal.ReadInt32(pointer);
                        }
                        else
                        {
                            handleCount = Marshal.ReadInt64(pointer);
                        }

                        int offset = IntPtr.Size;

                        int size = Marshal.SizeOf(typeof(SystemHandleEntry));

                        for(int i = 0; i < handleCount; i++)
                        {
                            SystemHandleEntry handleEntry = (SystemHandleEntry)Marshal.PtrToStructure((IntPtr)(pointer.ToInt64() + offset), typeof(SystemHandleEntry));

                            IntPtr newPointer = new IntPtr(pointer.ToInt64() + offset);

                            int newProcessID = Marshal.ReadInt32(newPointer);

                            if(handleEntry.OwnerProcessID == processID)
                            {
                                IntPtr handle = (IntPtr)handleEntry.HandleValue;

                                SystemHandleType handleType;

                                if(OpenFileHelper.GetHandleType(handle, handleEntry.OwnerProcessID, out handleType) == true)
                                {
                                    string devicePath;

                                    if(OpenFileHelper.GetFileNameFromHandle(handle, handleEntry.OwnerProcessID, out devicePath))
                                    {
                                        HandleInfo info = new HandleInfo();

                                        info.Name = devicePath;
                                        info.Type = handleType;

                                        yield return info;
                                    }
                                }
                            }

                            offset += size;
                        }
                    }
                }
                finally
                {
                    Marshal.FreeHGlobal(pointer);
                }
            }
            while(status == NTStatus.InfoLengthMismatch);
        }

        #endregion
    }
}