﻿using Microsoft.Win32.SafeHandles;
using System;
using System.Security.Permissions;

namespace TestProject
{
    /// <summary>
    /// 안전 객체 핸들
    /// </summary>
    [SecurityPermission(SecurityAction.LinkDemand, UnmanagedCode = true)]
    public sealed class SafeObjectHandle : SafeHandleZeroOrMinusOneIsInvalid
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - SafeObjectHandle(preExistingHandle, ownsHandle)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="preExistingHandle">사전 존재 핸들</param>
        /// <param name="ownsHandle">핸들 소유 여부</param>
        public SafeObjectHandle(IntPtr preExistingHandle, bool ownsHandle) : base(ownsHandle)
        {
            base.SetHandle(preExistingHandle);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 생성자 - SafeObjectHandle()

        /// <summary>
        /// 생성자
        /// </summary>
        private SafeObjectHandle() : base(true)
        {
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 핸들 해제하기 - ReleaseHandle()

        /// <summary>
        /// 핸들 해제하기
        /// </summary>
        /// <returns>처리 결과</returns>
        protected override bool ReleaseHandle()
        {
            return APIHelper.CloseHandle(base.handle);
        }

        #endregion
    }
}